<?php
/**
 * Plugin Name: RSS Feed Reader by Enebrus Kem Lem
 * Plugin URI: https://enebruskemlem.com/wordpress-plugins/rss-feed-reader
 * Description: Reads RSS Feeds automatically and publish their posts in your site linking the original site.
 * Version: 1.0.7
 * Author: Enebrus Kem Lem C.A.
 * Author URI: http://enebruskemlem.com
 * License: GPL2
 * Text Domain: ekl_rss_feed_reader
 * Domain Path: /languages
 */

// file is being called directly.
if (false === defined('ABSPATH')) {
  exit;
}

define('EKL_RSS_VERSION', 107);
define('EKL_RSS_CHECK_URL_ACTION', 'ekl_rss_check_rss_url');
define('EKL_RSS_UPDATE_OLD_TYPE_POSTS_ACTION', 'ekl_rss_update_old_type_posts');

/**
 * Actions to be performed when plugin is activated, those are:
 * - schedule the hourly job for check RSS Feed Sources
 *
 * @return void
 */
function ekl_rss_activate_plugin() : void
{
  // schedule the hourly RSS feed sources checking and logs the error in case of failure
  if (false === wp_schedule_event(ekl_rss_get_current_oclock_hour_timestamp(), 'hourly', 'ekl_rss_process_rss_sources_schedule')) {
    error_log('RSS Feed Sources processor job could not be scheduled');
  }
}

/**
 * Registers the metaboxes fields for the RSS sources
 *
 * @return void
 */
function ekl_rss_add_rss_sources_metaboxes() : void
{
  // adds the RSS fields
  add_meta_box(
    'rss_metaboxes',
    __('RSS feed data', 'ekl_rss_feed_reader'),
    'ekl_rss_show_rss_metaboxes_html',
    'ekl-rss-rss-source',
    'normal',
    'high'
  );
}

/**
 * Registers the RSS sources to be used for auto-feeding
 *
 * @return void
 */
function ekl_rss_add_rss_sources_post_type() : void
{
  // adds the RSS sources management
  register_post_type(
    'ekl-rss-rss-source',
    [
      'labels' => [
        'name'          => __('RSS sources', 'ekl_rss_feed_reader'),
        'singular_name' => __('RSS source', 'ekl_rss_feed_reader')
      ],
      'exclude_from_search' => false,
      'show_ui'             => true,
      'show_in_menu'        => true,
      'show_in_nav_menus'   => false,
      'show_in_admin_bar'   => false,
      'show_in_rest'        => false,
      'menu_icon'           => 'dashicons-rss',
      'supports'            => ['title', 'thumbnail', 'custom-fields'],
      'rewrite'             => ['slug' => 'rss-source']
    ]
  );
}

/**
 * Adds the menu to manually perform required plugin update actions
 *
 * @return void
 */
function ekl_rss_add_upgrade_menu() : void
{
  // adds the meny entry
  add_submenu_page(
    'edit.php?post_type=ekl-rss-rss-source',
    __('Upgrade Actions', 'ekl_rss_feed_reader'),
    __('Upgrade', 'ekl_rss_feed_reader' ),
    'manage_options',
    'ekl-rss-upgrade',
    'ekl_rss_show_upgrade_page'
  );
}

/**
 * Perform the required actions when deactivate the plugin:
 * - Removes the RSS Feed Source processor job
 *
 * @return void
 */
function ekl_rss_deactivate_plugin() : void
{
  // RSS Feed Sources processor job was scheduled
  if (wp_get_scheduled_event('ekl_rss_process_rss_sources_schedule')) {
    // removes the job schedule
    wp_unschedule_hook('ekl_rss_process_rss_sources_schedule');
  }
}

/**
 * Add common plugin script.
 *
 * @return void
 */
function ekl_rss_enqueue_scripts() : void
{
  // it is one page in admin interface.
  if (is_admin()) {
    return;
  }

  // add general script.
  wp_enqueue_script(
    'ekl-rss-feed-reader-common',
    plugin_dir_url(__FILE__).'assets/ekl-rss-feed-reader-common.js',
    [],
    EKL_RSS_VERSION,
    true
  );
}

/**
 * Handles the request for check a RSS feed URL
 *
 * @return void
 */
function ekl_rss_handle_check_rss_url_request() : void
{
  require_once ABSPATH.WPINC.'/class-simplepie.php';

  try {
    // nonce code is invalid or older than 12 hours
    if (wp_verify_nonce($_POST['nonce'], EKL_RSS_CHECK_URL_ACTION) !== 1) {
      throw new Exception(__('Wrong nonce code', 'ekl_rss_feed_reader'), __LINE__);
    }

    // URL does not have the correct format
    $url = filter_input(INPUT_POST, 'url', FILTER_VALIDATE_URL);
    if (empty($url)) {
      throw new Exception(__('Wrong url', 'ekl_rss_feed_reader'), __LINE__);
    }

    // gets the RSS content
    $feed = new SimplePie();
    $feed->set_feed_url($url);
    $feed->set_timeout(20);
    $feed->enable_cache(false);

    // there was an error during request
    if (false === $feed->init()) {
      throw new Exception(sprintf(__('Cannot check url %s', 'ekl_rss_feed_reader'), $url), __LINE__);
    }

    wp_send_json_success();
  }
  catch (Exception $ex) {
    wp_send_json_error(new WP_Error($ex->getCode(), $ex->getMessage()), 400);
  }
}

/**
 * Handles the request to update the old type posts
 *
 * @return void
 */
function ekl_rss_handle_update_old_type_posts_request() : void
{
  try {
    // nonce code is invalid or older than 12 hours
    if (wp_verify_nonce($_POST['nonce'], EKL_RSS_UPDATE_OLD_TYPE_POSTS_ACTION) !== 1) {
      throw new Exception(__('Wrong nonce code', 'ekl_rss_feed_reader'), __LINE__);
    }

    // get all old type posts
    $posts = get_posts([
      'post_type'      => 'dhm-rss-source',
      'posts_per_page' => -1,
      'post_status'    => array_keys(get_post_statuses())
    ]);

    // iterates over each post to update its type
    foreach ($posts as $post) {
      // update the post type
      $result = wp_update_post([
        'ID'        => $post->ID,
        'post_type' => 'ekl-rss-rss-source'
      ]);

      // post type could not be updated
      if ($result instanceof WP_Error) {
        throw new Exception($result->get_error_message(), __LINE__);
      }
    }

    wp_send_json_success();
  }
  catch (Exception $ex) {
    wp_send_json_error(new WP_Error($ex->getCode(), $ex->getMessage()), 400);
  }
}

/**
 * Tells if a RSS Feed Item already exists
 *
 * @param string $item_id Entry ID
 *
 * @return bool
 */
function ekl_rss_exists_rss_item(string $item_id) : bool
{
  // search the post by its feed ID
  $posts = get_posts([
    'posts_per_page' => 1,
    'post_status'    => array_keys(get_post_statuses()),
    'meta_query' => [
      [
        'key'   => 'ekl_rss_rss_item_id',
        'value' => $item_id
      ]
    ]
  ]);

  return false === empty($posts);
}

/**
 * Gives the extractors that can be used to get the attributes content
 *
 * @return array
 */
function ekl_rss_get_attribute_extractors() : array
{
  return [
    'full'   => __('Whole content', 'ekl_rss_feed_reader'),
    'regexp' => __('Regular expression', 'ekl_rss_feed_reader')
  ];
}

/**
 * Gives the current hour timestamp setting to zero the minutes and seconds.
 *
 * @return int
 */
function ekl_rss_get_current_oclock_hour_timestamp() : int
{
  // gets the next o'clock hour (removes the seconds and minutes)
  $hour = current_time('timestamp', true);
  $hour -= $hour % 60;
  $hour -= $hour % 3600;

  return $hour;
}

/**
 * Gives the RSS Feed custom attributes content.
 *
 * @param int $post_id RSS feed post.
 * @param bool $give_sanitizers Whether attributes must reveal their sanitizers (TRUE) or their content (FALSE). Default
 * is FALSE.
 *
 * @return array
 */
function ekl_rss_get_post_rss_attributes(int $post_id, bool $give_sanitizers = false) : array
{
  // attribute sanitizers must be given
  if ($give_sanitizers) {
    $result = [
      'url'                                => FILTER_SANITIZE_URL,
      'entries_count'                      => FILTER_SANITIZE_NUMBER_INT,
      'entry_picture_source'               => FILTER_SANITIZE_STRING,
      'entry_picture_index'                => FILTER_SANITIZE_NUMBER_INT,
      'entry_picture_xpath'                => FILTER_SANITIZE_STRING,
      'update_interval'                    => FILTER_SANITIZE_STRING,
      'entry_picture_tag'                  => FILTER_SANITIZE_STRING,
      'entry_picture_tag_attribute'        => FILTER_SANITIZE_STRING,
      'entry_picture_attribute_extractor'  => FILTER_SANITIZE_STRING,
      'entry_picture_extractor_expression' => FILTER_SANITIZE_STRING,
      'entry_picture_extractor_payload'    => FILTER_SANITIZE_NUMBER_INT
    ];
  }
  // content must be given
  else {
    // gets all post meta fields.
    $custom_fields = get_post_meta($post_id);

    $result = [
      'url'                                => $custom_fields['ekl_rss_url'][0] ?? '',
      'entries_count'                      => $custom_fields['ekl_rss_entries_count'][0] ?? 5,
      'entry_picture_source'               => $custom_fields['ekl_rss_entry_picture_source'][0] ?? 'tag_index',
      'entry_picture_index'                => $custom_fields['ekl_rss_entry_picture_index'][0] ?? 1,
      'entry_picture_xpath'                => $custom_fields['ekl_rss_entry_picture_xpath'][0] ?? '',
      'update_interval'                    => $custom_fields['ekl_rss_update_interval'][0] ?? '1',
      'entry_picture_tag'                  => $custom_fields['ekl_rss_entry_picture_tag'][0] ?? 'img',
      'entry_picture_tag_attribute'        => $custom_fields['ekl_rss_entry_picture_tag_attribute'][0] ?? 'src',
      'entry_picture_attribute_extractor'  => $custom_fields['ekl_rss_entry_picture_attribute_extractor'][0] ?? 'full',
      'entry_picture_extractor_expression' => $custom_fields['ekl_rss_entry_picture_extractor_expression'][0] ?? '',
      'entry_picture_extractor_payload'    => $custom_fields['ekl_rss_entry_picture_extractor_payload'][0] ?? 1
    ];
  }

  return $result;
}

/**
 * Gives the sources that can be used to fetch thumbnail for RSS items
 *
 * @return array
 */
function ekl_rss_get_rss_item_image_sources() : array
{
  return [
    'tag_index' => __('Tag index', 'ekl_rss_feed_reader'),
    'xpath'     => __('Tag XPath', 'ekl_rss_feed_reader')
  ];
}

/**
 * Alters the posts permalink for RSS posts
 *
 * @param string $permalink Current permalink
 * @param WP_Post $post Post
 * @param bool $leavename Whether the name must be keept
 *
 * @return string
 */
function ekl_rss_get_rss_item_permalink(string $permalink, WP_Post $post, bool $leavename) : string
{
  // gets the RSS Item link
  $rss_item_link = get_post_meta($post->ID, 'ekl_rss_rss_item_link', true);

  return $rss_item_link ?: $permalink;
}

/**
 * Gives the custom thumbnail information for RSS Feed Items
 *
 * @param array|false $image Current thumbnail data, or FALSE if there is no data
 * @param int $attachment_id Current thumbnail ID
 * @param string|array $size Thumbnail size
 * @param bool $icon Whether the image should be treated as an icon. Default false.
 *
 * @return array|false
 */
function ekl_rss_get_rss_item_thumbnail($image, $attachment_id, $size, $icon)
{
  global $post;

  // there is a post, and
  // there is a picture, and
  // the current picture ID is the post thumbnail
  if ($post && $image && ($attachment_id === get_post_thumbnail_id($post))) {
    // gets the RSS item thumbnail URL
    $url = get_post_meta($post->ID, 'ekl_rss_rss_item_thumbnail', true);

    // post has a custom thumbnail URL
    if ($url) {
      $image[0] = $url;
    }
  }

  return $image;
}

/**
 * Gives the allowed RSS feed update intervals
 *
 * @return array
 */
function ekl_rss_get_rss_update_intervals() : array
{
  return [
    '1'  => __('Hourly', 'ekl_rss_feed_reader'),
    '2'  => __('Every 2 Hours', 'ekl_rss_feed_reader'),
    '4'  => __('Every 4 Hours', 'ekl_rss_feed_reader'),
    '6'  => __('Every 6 Hours', 'ekl_rss_feed_reader'),
    '8'  => __('Every 8 Hours', 'ekl_rss_feed_reader'),
    '10' => __('Every 10 Hours', 'ekl_rss_feed_reader'),
    '12' => __('Every 12 Hours', 'ekl_rss_feed_reader'),
    '24' => __('Daily', 'ekl_rss_feed_reader')
  ];
}

/**
 * Process one RSS feed source in order to get new entries
 *
 * @param WP_Post $post Post to process
 *
 * @return void
 */
function ekl_rss_process_rss_source(WP_Post $post) : void
{
  require_once ABSPATH.WPINC.'/class-simplepie.php';

  // gets the RSS source custom fields
  extract(ekl_rss_get_post_rss_attributes($post->ID));

  try {
    // RSS URL has not been set
    if (empty($url)) {
      throw new Exception("{$post->post_title} feed URL has not been set");
    }

    // post picture must be found using XPath but XPath is empty
    if (('background' === $entry_picture_source) && empty($entry_picture_xpath)) {
      throw new Exception("{$post->post_title} feed picture XPath has not been set");
    }

    // gets the RSS content
    $feed = new SimplePie();
    $feed->set_feed_url($url);
    $feed->set_timeout(20);
    $feed->enable_cache(false);
    $feed->init();

    // gets the feed link to use it as author
    $feed_author = parse_url($feed->get_link(), PHP_URL_HOST);

    // gets the feed thumbnail to use as feed item thumbnail
    $feed_thumbnail_id = get_post_thumbnail_id($post);

    // searches the feed user
    $users = new WP_User_Query([
      'search'         => $feed_author,
      'search_columns' => ['user_login']
    ]);
    $feed_user = reset($users->get_results());

    // user does not exists
    if (empty($feed_user)) {
      // creates the feed user
      $feed_user_id = wp_create_user(
        $feed_author,
        md5(time().mt_rand(10, 1000)),
        "{$feed_author}@".parse_url(get_home_url())['host']
      );

      // user could not be created
      if ($feed_user_id instanceof WP_Error) {
        throw new Exception($feed_user_id->get_error_message());
      }
    }
    // user exists
    else {
      $feed_user_id = $feed_user->ID;
    }

    // fetches the RSS items
    $items = $feed->get_items(0, $entries_count);

    // reverse the items order because if not the first post in array will be the oldest
    $items = array_reverse($items);

    // iterates over each feed item to publish the new ones
    foreach ($items as $item) {
      // gets the item ID
      $item_id = $item->get_id(true);

      // item was already published
      if (ekl_rss_exists_rss_item($item_id)) {
        continue;
      }

      // gets the item title
      $item_title = $item->get_title();

      // creates the post
      $inserted_post_id = wp_insert_post(
        [
          'post_author'    => $feed_user_id,
          'post_content'   => $item->get_content(),
          'post_title'     => $item_title,
          'post_excerpt'   => $item->get_description(),
          'post_status'    => 'publish',
          'comment_status' => 'closed',
          'ping_status'    => 'closed',
          'post_date_gmt'  => $item->get_gmdate('Y-m-d H:i:s')
        ],
        true
      );

      // there was an error creating the post
      if ($inserted_post_id instanceof WP_Error) {
        throw new Exception($inserted_post_id->get_error_message());
      }

      // sets the post RSS Feed item information
      update_post_meta($inserted_post_id, 'ekl_rss_rss_item_id', $item_id);
      update_post_meta($inserted_post_id, 'ekl_rss_rss_item_link', $item->get_link());

      // sets the post thumbnail same as feed thumbnail (real picture is gotten via filter on show-time)
      set_post_thumbnail($inserted_post_id, $feed_thumbnail_id);

      // gets the post content
      $response = wp_remote_request($item->get_link(), ['timeout' => 20]);

      // there was an error
      if ($response instanceof WP_Error) {
        throw new Exception($response->get_error_message());
      }

      // parses the post content to search the thumbnail      
      $document                     = new DOMDocument();
      $document->preserveWhiteSpace = false;
      @$document->loadHTML($response['body']);

      // image is searched by XPath
      $value = '';
      if ('xpath' === $entry_picture_source) {
        // extracts the content image
        $xpath = new DOMXPath($document);
        $tags  = $xpath->query($entry_picture_xpath);

        // tag was found
        if (isset($tags[0])) {
          $value = $tags[0]->attributes->getNamedItem($entry_picture_tag_attribute)->nodeValue;
        }
      }
      // image is searched by tag index
      else {
        // get all given tags
        $tags = $document->getElementsByTagName($entry_picture_tag);

        // there were tags found
        if ($tags->length) {
          $tag_index = (($tags->length >= $entry_picture_index) ? $entry_picture_index : $tags->length) - 1;
          $value     = $tags->item($tag_index)->getAttribute($entry_picture_tag_attribute);
        }
      }

      // content must be extracted as is
      if ('full' === $entry_picture_attribute_extractor) {
        // image URL is the whole text
        $image_url = $value;
      }
      // content must be extracted using regular expression
      else {
        // extracts the attribute content
        $matches                            = [];
        $entry_picture_extractor_expression = htmlspecialchars_decode(
          html_entity_decode(
            $entry_picture_extractor_expression,
            ENT_QUOTES
          ),
          ENT_QUOTES
        );
        preg_match('/'.$entry_picture_extractor_expression.'/', $value, $matches);

        // image URL was found
        if (false === empty($matches[$entry_picture_extractor_payload])) {
          $image_url = $matches[$entry_picture_extractor_payload];
        }
      }

      // image was found
      if (false === empty($image_url)) {
        // sets the post RSS Feed item picture URL
        update_post_meta(
          $inserted_post_id,
          'ekl_rss_rss_item_thumbnail',
          strtok(trim($image_url, " \t\n\r\0\x0B\"'"), '?')
        );
      }
    }
  }
  catch (Exception $ex) {
    error_log($ex->getMessage());
  }
}

/**
 * Processes the RSS Feed Sources according with their scheduled update frequency
 *
 * @return void
 */
function ekl_rss_process_rss_sources_schedule() : void
{
  // gets the current hour
  $hour = date('H', ekl_rss_get_current_oclock_hour_timestamp());

  error_log('Running RSS Feed Source processing at '.date('d/m/Y H:i:s')." for UTC hour {$hour}");

  // gets the update intervals
  $intervals = array_keys(ekl_rss_get_rss_update_intervals());
  sort($intervals, SORT_NUMERIC);

  // keeps only intervals that must run
  $intervals = array_filter($intervals, function ($interval) use ($hour) {
    return 0 === $hour % $interval;
  });

  error_log('RSS update intervals to be used are: '.implode(', ', $intervals));

  // query all RSS Feed Sources matching allowed intervals
  $posts = get_posts([
    'posts_per_page' => -1,
    'post_status'    => 'publish',
    'post_type'      => 'ekl-rss-rss-source',
    'meta_query'     => [
      [
        'key'     => 'ekl_rss_update_interval',
        'value'   => $intervals,
        'compare' => 'IN'
      ]
    ]
  ]);

  error_log('Number of RSS Feed Sources to process are: '.count($posts));

  // iterates over each RSS Feed Source to process them
  foreach ($posts as $post) {
    error_log("Processing {$post->post_title} RSS Feed Source");

    // processes the RSS Feed Source
    ekl_rss_process_rss_source($post);

    error_log("Processed {$post->post_title} RSS Feed Source");
  }

  error_log("RSS Feed Sources were processed");
}

/**
 * Stores the custom data for the RSS source post type
 *
 * @param int $post_id Post ID
 * @param WP_Post $post Post data
 * @param bool $update TRUE if post is being updated, FALSE otherwise
 *
 * @return void
 */
function ekl_rss_save_rss_source(int $post_id, WP_Post $post, bool $update) : void
{
  // form is being autosaved.
  if (defined('DOING_AUTOSAVE') && DOING_AUTOSAVE) {
    return;
  }

  // user does not have permission.
  if (false === current_user_can('edit_post', $post_id)) {
    return;
  }

  // set custom fields for RSS source.
  $fields = ekl_rss_get_post_rss_attributes(0, true);

  // iterate over each RSS field to get its data.
  foreach ($fields as $field => $sanitizer) {
    // set field prefix.
    $field = "ekl_rss_{$field}";

    // get field content.
    $content = addslashes(filter_input(INPUT_POST, $field, $sanitizer));

    // field is empty.
    if (empty($content)) {
      delete_post_meta($post_id, $field);
    }
    // field is not empty.
    else {
      update_post_meta($post_id, $field, $content);
    }
  }

  // RSS fetching must run inmediately
  if ((false === empty($_POST['ekl_rss_run_on_save'])) && 'yes' === $_POST['ekl_rss_run_on_save']) {
    // processes the rss feed source
    ekl_rss_process_rss_source($post);
  }
}

/**
 * Shows the RSS source custom fields metaboxes
 *
 * @return void
 */
function ekl_rss_show_rss_metaboxes_html() : void
{
  global $post;

  // gets the custom fields
  extract(ekl_rss_get_post_rss_attributes($post->ID));

  require __DIR__.'/admin/rss/metaboxes.php';
}

/**
 * Shows the page used to perform custom upgrade actions
 *
 * @return void
 */
function ekl_rss_show_upgrade_page() : void
{
  // gets old type posts
  $old_type_posts = get_posts([
    'post_type'      => 'dhm-rss-source',
    'posts_per_page' => -1,
    'post_status'    => array_keys(get_post_statuses())
  ]);

  // count posts using the old post type
  $old_type_posts_count = count($old_type_posts);

  require __DIR__.'/admin/rss/upgrade.php';
}

// sets the actions to perform on plugin activation
register_activation_hook(__FILE__, 'ekl_rss_activate_plugin');
register_deactivation_hook(__FILE__, 'ekl_rss_deactivate_plugin');

// adds the RSS settings page
add_action('init', 'ekl_rss_add_rss_sources_post_type', 10, 0);
add_action('add_meta_boxes_ekl-rss-rss-source', 'ekl_rss_add_rss_sources_metaboxes', 10, 0);
add_action('wp_ajax_ekl_rss_check_rss_url', 'ekl_rss_handle_check_rss_url_request', 10, 0);
add_action('save_post_ekl-rss-rss-source', 'ekl_rss_save_rss_source', 10, 3);
add_filter('post_link', 'ekl_rss_get_rss_item_permalink', 10, 3);
add_filter('wp_get_attachment_image_src', 'ekl_rss_get_rss_item_thumbnail', 10, 4);
add_action('ekl_rss_process_rss_sources_schedule', 'ekl_rss_process_rss_sources_schedule', 10, 0);
add_action('admin_menu', 'ekl_rss_add_upgrade_menu', 10, 0);
add_action('wp_ajax_ekl_rss_update_old_type_posts', 'ekl_rss_handle_update_old_type_posts_request', 10, 0);
add_action('wp_enqueue_scripts', 'ekl_rss_enqueue_scripts', 10, 0);